def compute_epoch_loss(trainer):
    train_loss = 0
    eval_loss = 0

    for s in trainer.state.train.minibatch_stats:
        train_loss += s['loss']

    for s in trainer.state.eval.minibatch_stats:
        eval_loss += s['loss']

    trainer.state.train_loss = train_loss
    trainer.state.eval_loss = eval_loss

def compute_epoch_acc(trainer):
    n = 0
    correct_predictions = 0
    for s in trainer.state.train.minibatch_stats:
        correct_predictions += s['correct_predictions']
        n += s['size']
    train_acc = correct_predictions / n

    n = 0
    correct_predictions = 0
    for s in trainer.state.eval.minibatch_stats:
        correct_predictions += s['correct_predictions']
        n += s['size']
    eval_acc = correct_predictions / n

    trainer.state.train_acc = train_acc
    trainer.state.eval_acc = eval_acc



import torch
import torch.nn.functional as F

from torchfuel.trainers.generic import GenericTrainer
import torchfuel.trainers.const as const
from torchfuel.trainers.metrics import compute_epoch_acc

class BasicClassificationTrainer(GenericTrainer):
    def __init__(self, device, model, optimiser, scheduler,
                 model_name='model.pt', print_perf=True):
        super().__init__(
            device,
            model,
            optimiser,
            scheduler,
            model_name=model_name,
            print_perf=print_perf
        )

        self._hooks[const.AFTER_EPOCH].append(compute_epoch_acc)

    def compute_loss(self, output, y):
        return F.cross_entropy(output, y)

    def print_epoch_performance(self, epoch, train_epoch_stats, eval_epoch_stats):
        train_loss = self.state.train_loss
        eval_loss = self.state.eval_loss

        train_acc = self.state.train_acc
        eval_acc = self.state.eval_acc

        elapsed_time = self.state.general.elapsed_time

        s = ('(Epoch #{}) Train loss {:.3f} & acc {:.2f}'
             ' | Eval loss {:.4f} & acc {:.2f} ({:.2f} s)')

        s = s.format(epoch, train_loss, train_acc,
                     eval_loss, eval_acc, elapsed_time)
        print(s)

    def update_best_model(self, best_model, eval_epoch_stats):
        eval_acc = self.state.eval_acc
        if best_model is None or eval_acc < best_model['acc']:
            best_model = {}
            best_model['acc'] = eval_acc
            best_model['model'] = self.model.state_dict()

        return best_model

    def compute_correct_preds(self, output, y):
        _, pred = torch.max(output, 1)
        correct = torch.sum(pred == y).item()
        return correct

    def compute_minibatch_statistics(self, X, y, output, loss):
        d = super().compute_minibatch_statistics(X, y, output, loss)
        correct = self.compute_correct_preds(output, y)
        batch_size = X.size(0)
        d.update({'correct_predictions': correct, 'size': batch_size})
        return d



import time


def log_start_time(trainer):
    trainer.state.general.start_time = time.time()


def compute_epoch_time(trainer):
    end_time = time.time()
    trainer.state.general.elapsed_time = end_time - trainer.state.general.start_time


def step_scheduler(trainer):
    trainer.scheduler.step()


def step_on_plateau_scheduler(trainer):
    trainer.scheduler.step(trainer.state.eval_loss)



import pickle
import time
from abc import abstractmethod
from functools import wraps

import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from tqdm import tqdm

import torchfuel.trainers.const as const
from torchfuel.trainers.generic_hooks import (compute_epoch_time,
                                              log_start_time,
                                              step_on_plateau_scheduler,
                                              step_scheduler)
from torchfuel.trainers.metrics import compute_epoch_loss
from torchfuel.utils.time_parser import parse_seconds


class Placeholder:
    def __init__(self):
        self.stored_objects = {}

    def __repr__(self):
        arg = ','.join(self.stored_objects.keys())
        return 'Placeholder with objects ({})'.format(arg)

    def __setattr__(self, name, value):
        if name != 'stored_objects':
            self.stored_objects[name] = value
        else:
            super().__setattr__(name, value)

    def __getattr__(self, name):
        '''
        intercept lookups which would raise an exception
        to check if variable is being stored
        '''
        if name in self.stored_objects:
            return self.stored_objects[name]
        else:
            return super().__getattr__(name)

    def pickle_safe(self, file):
        pickle.dump(self, open(file, 'wb'))


class State:
    def __init__(self):
        self.general = Placeholder()
        self.train = Placeholder()
        self.eval = Placeholder()


class GenericTrainer:
    def __init__(self, device, model, optimiser, scheduler,
                 post_epoch_hooks=None,
                 model_name='model.pt', print_perf=True):
        self.device = device
        self.model = model
        self.optimiser = optimiser
        self.scheduler = scheduler
        self.model_name = model_name

        self.state = State()

        self._hooks = {c: list() for c in [const.AFTER_EPOCH, const.BEFORE_EPOCH,
                                           const.AFTER_MINIBATCH, const.BEFORE_MINIBATCH,
                                           const.AFTER_TRAIN, const.BEFORE_TRAIN,
                                           const.AFTER_EVAL, const.BEFORE_EVAL]}

        self._hooks[const.BEFORE_EPOCH].append(log_start_time)
        self._hooks[const.AFTER_EPOCH].append(compute_epoch_loss)
        self._hooks[const.AFTER_EPOCH].append(compute_epoch_time)

        self.print_perf = print_perf
        if print_perf:
            self.print_template = ()

        if isinstance(self.scheduler, optim.lr_scheduler.ReduceLROnPlateau):
            self._hooks[const.AFTER_EPOCH].append(step_on_plateau_scheduler)
        else:
            self._hooks[const.BEFORE_EPOCH].append(step_scheduler)

    @abstractmethod
    def compute_loss(self, output, y):
        pass

    def execute_on(self, where):
        def wrapper(func):
            if where in self._hooks:
                self._hooks[where].append(func)
            else:
                raise ValueError('Hook {} does not exists'.format(where))
            return func
        return wrapper

    def run_hooks(self, where):
        if where in self._hooks:
            for hook in self._hooks[where]:
                hook(self)
        else:
            raise ValueError('Hook {} does not exists'.format(where))

    def compute_minibatch_statistics(self, X, y, output, loss):
        return {'loss': loss}

    def print_epoch_performance(self, epoch, train_epoch_stats, eval_epoch_stats):
        train_loss = self.state.train_loss
        eval_loss = self.state.eval_loss
        elapsed_time = self.state.general.elapsed_time

        s = ('(Epoch #{}) Train loss {:.3f}'
             ' | Eval loss {:.4f} ({:.2f} s)')

        s = s.format(epoch, train_loss,
                     eval_loss, elapsed_time)
        print(s)

    def update_best_model(self, best_model, eval_epoch_stats):
        eval_loss = self.state.eval_loss
        if best_model is None or eval_loss < best_model['loss']:
            best_model = {}
            best_model['loss'] = eval_loss
            best_model['model'] = self.model.state_dict()

        return best_model

    def train_minibatch(self, X, y):
        output = self.model(X)
        loss = self.compute_loss(output, y)

        self.optimiser.zero_grad()
        loss.backward()
        self.optimiser.step()

        batch_size = X.size(0)
        total_loss = loss.item() * batch_size
        output = output.detach()
        return output, total_loss

    def eval_minibatch(self, X, y):
        output = self.model(X)
        loss = self.compute_loss(output, y)

        batch_size = X.size(0)
        total_loss = loss.item() * batch_size
        output = output.detach()
        return output, total_loss

    def train_epoch(self, dataloader, epoch):
        self.model.train()

        msg = 'Training (epoch {})'.format(epoch)
        self.state.train.minibatch_stats = []

        for X, y in tqdm(dataloader, desc=msg, leave=False):
            X = X.to(self.device)
            y = y.to(self.device)

            output, loss = self.train_minibatch(X, y)
            minibatch_stats = self.compute_minibatch_statistics(X, y, output, loss)
            self.state.train.minibatch_stats.append(minibatch_stats)

    def eval_epoch(self, dataloader, epoch):
        self.model.eval()

        msg = 'Evaluating (epoch {})'.format(epoch)
        self.state.eval.minibatch_stats = []

        with torch.set_grad_enabled(False):
            for X, y in tqdm(dataloader, desc=msg, leave=False):
                X = X.to(self.device)
                y = y.to(self.device)

                output, loss = self.eval_minibatch(X, y)
                minibatch_stats = self.compute_minibatch_statistics(X, y, output, loss)
                self.state.eval.minibatch_stats.append(minibatch_stats)

    def save_model(self, epoch, best_model):
        torch.save({
            'epoch': epoch,
            'model_state': self.model.state_dict(),
            'optimiser_state': self.optimiser.state_dict(),
            'scheduler_state': self.scheduler.state_dict(),
            'best_model': best_model,
        }, self.model_name)

    def load_model(self):
        checkpoint = torch.load(self.model_name)
        self.model.load_state_dict(checkpoint['model_state'])
        self.optimiser.load_state_dict(checkpoint['optimiser_state'])
        self.scheduler.load_state_dict(checkpoint['scheduler_state'])

        start_epoch = checkpoint['epoch']
        best_model = checkpoint['best_model']

        return start_epoch, best_model

    def fit(self, epochs, train_dataloader, eval_dataloader):
        try:
            start_epoch, best_model = self.load_model()
            # file does not exist or pytorch error (model architecture changed)
        except:
            start_epoch = 0
            best_model = None

        for epoch in range(start_epoch, epochs):
            # run hooks before epoch
            self.run_hooks(const.BEFORE_EPOCH)

            # run hooks before train
            self.run_hooks(const.BEFORE_TRAIN)

            train_epoch_stats = self.train_epoch(train_dataloader, epoch)

            # run hooks after train
            self.run_hooks(const.AFTER_TRAIN)

            # run hooks before eval
            self.run_hooks(const.BEFORE_EVAL)

            eval_epoch_stats = self.eval_epoch(eval_dataloader, epoch)

            # run hooks after eval
            self.run_hooks(const.AFTER_EVAL)

            # run hooks after epoch
            self.run_hooks(const.AFTER_EPOCH)

            best_model = self.update_best_model(best_model, eval_epoch_stats)

            if self.print_perf:
                self.print_epoch_performance(epoch, train_epoch_stats, eval_epoch_stats)

            if epoch != 0:
                self.save_model(epoch, best_model)

        end_time = time.time()
        elapsed_time = end_time - self.state.general.start_time
        days, hours, minutes, seconds = parse_seconds(elapsed_time)
        print('Training done in {}d, {}h {}min {:.2f}s'.format(days, hours, minutes, seconds))

        return self.model.load_state_dict(best_model['model'])



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    requirements = [p.strip() for p in f.readlines()]

with open('README.md') as f:
    long_description = f.read()


setup(
    name='torchfuel',
    version='0.1.0',
    description='Library build on top of pytorch to fuel productivity',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/vturrisi/torchfuel',
    author='Victor Turrisi',
    author_email='vt.turrisi@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import torch


class GaussianNoiser():
    def __init__(self, noise_amount):
        self.noise_amount = noise_amount

    def __call__(self, imgs):
        device = imgs.device
        noisy_imgs = imgs.clone()

        # gaussian noise
        noise = torch.rand(noisy_imgs.size()).to(device)

        noisy_imgs += self.noise_amount * noise

        # clamp between 0 and 1
        noisy_imgs.clamp_(0, 1)

        return noisy_imgs


class DropPixelNoiser():
    def __init__(self, noise_chance):
        self.noise_chance = noise_chance

    def __call__(self, imgs):
        noisy_imgs = imgs.clone()
        # noise described in the denoising autoencoder paper from bengio
        for img in noisy_imgs:
            for i in range(img.size(1)):
                for j in range(img.size(2)):
                    if torch.rand(1).item() < self.noise_chance:
                        img[:, i, j] = 0
        return noisy_imgs



import torch
from torch.utils.data import DataLoader, Dataset
from torchvision import datasets, models, transforms


class ImageDataLoader:
    def __init__(self,
                 train_data_folder: str, eval_data_folder: str,
                 pil_transformations: list = None,
                 tensor_transformations: list = None,
                 batch_size=64,
                 shuffle=True,
                 num_workers=4,
                 imagenet_format=True,
                 size=None, mean: list = None, std: list = None,
                 apply_transforms_to_eval: bool = False,
                 dataset_class: Dataset = datasets.ImageFolder):

        if pil_transformations is None:
            pil_transformations = []

        if tensor_transformations is None:
            tensor_transformations = []

        self.train_data_folder = train_data_folder
        self.eval_data_folder = eval_data_folder
        self.pil_transformations = pil_transformations
        self.tensor_transformations = tensor_transformations
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.num_workers = num_workers
        self.imagenet_format = imagenet_format
        self.size = size
        self.mean = mean
        self.std = std
        self.apply_transforms_to_eval = apply_transforms_to_eval

        self.dataset_class = dataset_class

    def prepare(self):
        if self.imagenet_format:
            assert all(v is None for v in [self.size, self.mean, self.std])

            size = 224
            mean = [0.485, 0.456, 0.406]
            std = [0.229, 0.224, 0.225]
            normalise = True
        elif all(v is not None for v in [self.size, self.mean, self.std]):
            size = self.size
            mean = self.mean
            std = self.std
            normalise = False
        elif self.size is not None:
            size = self.size
            normalise = False
        else:
            raise Exception('imagenet_format should be True or size must be specified')

        t = [
            transforms.Resize(size),
            *self.pil_transformations,
            transforms.ToTensor(),
        ]
        if normalise:
            t.append(transforms.Normalize(mean, std))
        t.extend(self.tensor_transformations)

        train_transforms = transforms.Compose(t)

        if self.apply_transforms_to_eval:
            eval_transforms = train_transforms
        else:
            t = [
                transforms.Resize(size),
                transforms.ToTensor(),
            ]
            if normalise:
                t.append(transforms.Normalize(mean, std))
            eval_transforms = transforms.Compose(t)

        data_transforms = {
            'train': train_transforms,
            'eval': eval_transforms,
        }

        train_data = self.dataset_class(self.train_data_folder, data_transforms['train'])
        train_dataloader = DataLoader(train_data, self.batch_size,
                                      shuffle=self.shuffle, num_workers=self.num_workers)

        eval_data = self.dataset_class(self.eval_data_folder, data_transforms['eval'])
        eval_dataloader = DataLoader(eval_data, self.batch_size,
                                     shuffle=False, num_workers=self.num_workers)

        class_names = train_data.classes
        n_classes = len(class_names)

        return train_dataloader, eval_dataloader, n_classes


class ImageToImageReader(datasets.ImageFolder):
        def __getitem__(self, index):
            img, y = super().__getitem__(index)
            return img, img


class ImageToImageDataLoader(ImageDataLoader):
    def __init__(self,
                 train_data_folder: str, eval_data_folder: str,
                 pil_transformations: list = None,
                 tensor_transformations: list = None,
                 batch_size=64,
                 shuffle=True,
                 num_workers=4,
                 imagenet_format=True,
                 size=None, mean: list = None, std: list = None,
                 apply_transforms_to_eval: bool = False):

        super().__init__(
            train_data_folder, eval_data_folder,
            pil_transformations,
            tensor_transformations,
            batch_size,
            shuffle,
            num_workers,
            imagenet_format,
            size,
            mean,
            std,
            apply_transforms_to_eval,
            dataset_class=ImageToImageReader
        )

    def prepare(self):
        train_dataloader, eval_dataloader, _ = super().prepare()
        return train_dataloader, eval_dataloader



import torch.nn.functional as F

from torchfuel.trainers.generic import GenericTrainer


class BasicAutoencoderTrainer(GenericTrainer):
    def __init__(self, device, model, optimiser, scheduler,
                 model_name='model.pt', print_perf=True):
        super().__init__(
            device,
            model,
            optimiser,
            scheduler,
            model_name=model_name,
            print_perf=print_perf
        )

    def compute_loss(self, output, y):
        return F.mse_loss(output, y)



import torch.nn as nn


class Flatten(nn.Module):
    def forward(self, imgs):
        return imgs.view(-1, 56 * 56 * 3)


class ReshapeToImg(nn.Module):
    def forward(self, imgs):
        return imgs.view(-1, 3, 56, 56)
